package org.processmining.humbertogarciacaballero.models.model.layout.stage.sequencer;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;

import org.processmining.humbertogarciacaballero.models.model.PNConstants;
import org.processmining.humbertogarciacaballero.models.model.layout.data.LayerContainer;
import org.processmining.humbertogarciacaballero.models.model.layout.stage.HierarchicalLayoutContext;

import prefuse.data.Graph;
import prefuse.data.Node;

public class HeuristicCrossingSequencer extends Sequencer {

    public HeuristicCrossingSequencer(HierarchicalLayoutContext context) {
        super(context);
    }

    /**
     * Performs a vertex ordering within ranks as described by Gansner et al
     * 1993
     */
    public int run(Graph graph) {
        boolean direction = true;
        LayerContainer[] bestIteration = new LayerContainer[2];
        int[] bestCrossingIteration = new int[2];

        for (int d = 0; d < 2; d++) {
            LayerContainer order = initOrder(graph, direction);
            bestIteration[d] = new LayerContainer(order);
            bestCrossingIteration[d] = crossing(bestIteration[d]);

            for (int i = 0; i < getContext().getMaxIterations(); i++) {
                weightedMedian(order, i, i % 4 > 2);
                transpose(order, i, i % 4 > 2);

                int currentCrossing = crossing(order);

                if (currentCrossing < bestCrossingIteration[d]) {
                    bestIteration[d] = new LayerContainer(order);
                    bestCrossingIteration[d] = currentCrossing;
                }
            }

            unvisitNodes(graph);
            direction = !direction;
        }

        LayerContainer best;
        if (bestCrossingIteration[0] < bestCrossingIteration[1]) {
            best = bestIteration[0];
        } else {
            best = bestIteration[1];
        }

        getContext().getLayerContainer().reset(best);
        getContext().getLayerContainer().print();

        if (PNConstants.DEBUG) {
            s_logger.info("Crossings: " + crossing(best));
        }

        return 0;
    }

    private void transpose(LayerContainer order, int i, boolean exchangeOnEquality) {
        boolean downward = (i % 2 == 0);
        boolean improved = true;
        // Prevent infinite loop
        int count = 0;

        while (improved && count < 10) {
            improved = false;

            if (downward) {
                for (int j = 1; j < order.getSize(); j++) {
                    improved |= transpose(order, j, -1, exchangeOnEquality);
                }
            } else {
                for (int j = order.getSize() - 2; j >= 0; j--) {
                    improved |= transpose(order, j, +1, exchangeOnEquality);
                }
            }
            count++;
        }
    }

    private boolean transpose(LayerContainer order, int layer, int nextLayer, boolean exchangeOnEquality) {
        boolean improved = false;
        List<Node> layerNodes = order.getNodes(layer);
        List<Node> nextLayerNodes = order.getNodes(layer + nextLayer);

        int layerNodesSize = layerNodes.size();
        for (int k = 0; k < layerNodesSize - 1; k++) {
            Node v = layerNodes.get(k);
            Node w = layerNodes.get(k + 1);
            int c1 = calculateCrossings(layerNodes, nextLayerNodes, v, k, w, k + 1);
            int c2 = calculateCrossings(layerNodes, nextLayerNodes, w, k, v, k + 1);

            if (c1 > c2 || (c1 == c2 && exchangeOnEquality)) {
                improved = true;
                order.swap(layer, k + 1, k);
            }
        }

        return improved;
    }

    private void weightedMedian(LayerContainer order, int i, boolean exchangeOnEquality) {
        boolean downward = (i % 2 == 0);

        if (downward) {
            for (int j = 1; j < order.getSize(); j++) {
                medianRank(order, j, -1, exchangeOnEquality);
            }
        } else {
            for (int j = order.getSize() - 2; j >= 0; j--) {
                medianRank(order, j, +1, exchangeOnEquality);
            }
        }
    }

    private LayerContainer initOrder(Graph graph, boolean outDirection) {
        LayerContainer lc = getContext().getLayerContainer();
        LayerContainer order = new LayerContainer();

        unvisitNodes(graph);

        Queue<Node> queue = new LinkedList<>();
        Iterator<Node> iter;
        if (outDirection) {
            iter = getRootNodes(graph);
        } else {
            iter = getSinkNodes(graph);
        }

        while (iter.hasNext()) {
            Node node = iter.next();
            queue.add(node);
            setVisitedNode(node, true);
            order.addNode(node, lc.getLayer(node));
        }

        while (!queue.isEmpty()) {
            Node node = queue.poll();

            if (outDirection) {
                iter = getSuccessors(node, false);
            } else {
                iter = getRealPredecessors(node);
            }
            while (iter.hasNext()) {
                Node neighbour = iter.next();
                if (!isVisited(neighbour)) {
                    int layer = lc.getLayer(neighbour);
                    order.addNode(neighbour, layer);
                    setVisitedNode(neighbour, true);
                    queue.add(neighbour);
                }
            }
        }

        return order;
    }

    /**
     * Attempts to minimise the median placement of connected cells on this rank
     * and one of the adjacent ranks
     * 
     * @param rankValue
     *            the layer number of this rank
     * @param downwardSweep
     *            whether or not this is a downward sweep through the graph
     */
    private void medianRank(LayerContainer order, int layer, int nextLayer, boolean exchangeOnEquality) {
        List<Node> orderedNodes = new ArrayList<>(order.getNodes(layer));
        final Map<Node, Float> medians = new HashMap<>(orderedNodes.size());

        for (int i = 0; i < orderedNodes.size(); i++) {
            Node v = orderedNodes.get(i);
            medians.put(v, medianValue(order, v, nextLayer));
        }

        Collections.sort(orderedNodes, new Comparator<Node>() {
            @Override
            public int compare(Node o1, Node o2) {
                return Float.compare(medians.get(o1), medians.get(o2));
            }
        });

        List<Node> finalOrder = order.getNodes(layer);
        int currentPosition = 0;

        for (Node v : orderedNodes) {
            // If the median is ge 0 then we can place the node
            // otherwise it'd be -1 in which case it should remain
            // the same position
            if (medians.get(v) >= 0) {
                // if in this position there is a non-movable node we look for
                // another spot
                while (medians.get(finalOrder.get(currentPosition)) == -1) {
                    currentPosition++;
                }
                finalOrder.set(currentPosition++, v);
                getNodeParams(v).rank = currentPosition;
            }
        }
    }
}
