package org.processmining.humbertogarciacaballero.models.util;

import java.util.Iterator;

import prefuse.data.Edge;
import prefuse.data.Graph;
import prefuse.data.Node;
import prefuse.data.Table;
import prefuse.data.Tuple;
import prefuse.data.expression.Predicate;
import prefuse.data.expression.parser.ExpressionParser;
import prefuse.data.util.Sort;

public class GraphUtil {

    public static void copyColumns(Tuple source, Tuple target) {
        if (!target.getSchema().isAssignableFrom(source.getSchema())) {
            throw new IllegalArgumentException("Target's schema has to be assignable from source's schema");
        }
        for (int i = 0; i < source.getColumnCount(); i++) {
            target.set(i, source.get(i));
        }
    }

    public static String[] getColumnNames(Table t) {
        String[] columnNames = new String[t.getColumnCount()];
        for (int ii = 0; ii < t.getColumnCount(); ii++) {
            columnNames[ii] = t.getColumnName(ii);
        }
        return columnNames;
    }

    public static Tuple copyTuple(Tuple t) {
        Tuple tCopy = null;

        Table table = t.getTable().getSchema().instantiate();
        int row = table.addRow();
        tCopy = table.getTuple(row);

        for (int i = 0; i < t.getColumnCount(); i++) {
            tCopy.set(i, t.get(i));
        }

        return tCopy;
    }

    public static Table copyTable(Table t) {
        Table tCopy = new Table();
        tCopy.addColumns(t.getSchema());

        for (Iterator ii = t.tuples(); ii.hasNext();) {
            Tuple tuple = (Tuple) ii.next();
            tCopy.addTuple(tuple);
        }
        return tCopy;
    }

    public static Table copySparseTable(Table t) {
        Table tCopy = t.getSchema().instantiate();
        for (int ii = t.getMinimumRow(); ii < t.getMaximumRow() + 1; ii++) {
            Tuple tuple = t.getTuple(ii);
            if (tuple != null) {
                tCopy.addTuple(tuple);
            } else {
                tCopy.addRow(); // add an empty row
            }
        }
        return tCopy;
    }

    public static Graph copySparseGraph(Graph g) {
        Table nodeTableCopy = copySparseTable(g.getNodeTable());
        Table edgeTableCopy = copySparseTable(g.getEdgeTable());
        Graph gCopy = new Graph(nodeTableCopy, edgeTableCopy, g.isDirected(), g.getNodeKeyField(),
                g.getEdgeSourceField(), g.getEdgeTargetField());

        for (int i = 0; i < g.getNodeTable().getMaximumRow() + 1; i++) {
            if (!g.getNodeTable().isValidRow(i)) {
                gCopy.getNodeTable().removeRow(i);
            }
        }

        for (int i = 0; i < g.getEdgeTable().getMaximumRow() + 1; i++) {
            if (!g.getEdgeTable().isValidRow(i)) {
                gCopy.getEdgeTable().removeRow(i);
            }
        }

        return gCopy;
    }

    public static void printTable(Table t) {
        Iterator ii = t.tuples();
        while (ii.hasNext()) {
            System.out.println((Tuple) ii.next());
        }
    }

    public static void printGraph(Graph g) {
        Table nodeTable = g.getNodeTable();
        Table edgeTable = g.getEdgeTable();
        System.out.println("**Nodes**");
        printTable(nodeTable);
        System.out.println("**Edges**");
        printTable(edgeTable);
    }

    public static Table sortCopyByColumn(Table t, String sortAttribute, boolean ascending) {
        Sort sortByAttribute = new Sort(new String[] { sortAttribute }, new boolean[] { ascending });
        Predicate doNotFilter = ExpressionParser.predicate("TRUE");
        Table nodeTableColumnSorted = t.select(doNotFilter, sortByAttribute);
        return nodeTableColumnSorted;
    }

    // make a new graph by copying the node and edge tables
    public static Graph copyAndMakeGraph(Table nodeTable, Table edgeTable, boolean directed) {
        Table nodeTableCopy = copyTable(nodeTable);
        Table edgeTableCopy = copyTable(edgeTable);
        //////////////////
        Graph newGraph = new Graph(nodeTableCopy, edgeTableCopy, directed);
        return newGraph;
    }

    public static Table getSorted(Table t) {
        return getSortedByColumns(t, getColumnNames(t));
    }

    public static Table getSortedByColumns(Table t, String[] columnNames) {
        for (int ii = 0; ii < columnNames.length; ii++) {
            String columnName = columnNames[ii];
            if (t.getColumn(columnName) == null) {
                System.out.println("Cannot find column " + columnName);
            }
        }
        Sort tSort = new Sort(columnNames);
        Table sortedTable = t.select(ExpressionParser.predicate("TRUE"), tSort);
        return sortedTable;
    }

    public static boolean areEqual(Tuple tu1, Tuple tu2) {
        if (tu1.getColumnCount() != tu2.getColumnCount()) {
            return false;
        }

        for (int ii = 0; ii < tu1.getColumnCount(); ii++) {
            Object columnContents1 = tu1.get(ii);

            Object columnContents2 = null;
            boolean foundMatchingColumn = false;
            for (int kk = 0; kk < tu2.getColumnCount(); kk++) {

                if (tu2.getColumnName(kk).equals(tu1.getColumnName(ii))) {
                    columnContents2 = tu2.get(kk);
                    foundMatchingColumn = true;
                    break;
                }
            }

            // TODO: Possibly remove this, since it SHOULD be guaranteed
            // not to happen by a check run before this algorithm
            if (!foundMatchingColumn) {
                return false;
            }

            String columnName = tu1.getColumnName(ii);

            if (columnContents1 == null && columnContents2 == null) {
                // nulls are equal to each other!
                continue;
            } else if (columnContents1 == null) {
                return false;
            } else if (columnContents2 == null) {
                return false;
            } else if (!columnContents1.equals(columnContents2)) {
                // neither are null, but they are still not equal.
                return false;
            }
        }

        // all column contents are equal.
        return true;
    }
    
	/**
	 * Gets the first node with out degree equals to 0, that is, a sink node.
	 * 
	 * @return a sink node or null if there are no sink nodes.
	 */
	public static Node getSink(Graph graph) {
		Iterator<Node> iter = graph.nodes();

		while (iter.hasNext()) {
			Node n = iter.next();
			if (n.getOutDegree() == 0)
				return n;
		}

		return null;
	}
	
	/**
	 * Gets the first node with in degree equals to 0, that is, a source node.
	 * 
	 * @return a source node or null if there are no source nodes.
	 */
    public static Node getSource(Graph graph) {
    	Iterator<Node> iter = graph.nodes();
    	
    	while (iter.hasNext()) {
    		Node n = iter.next();
    		if (n.getInDegree() == 0)
    			return n;
    	}
    	
    	return null;
    }
    
	/**
	 * Gets the node with the maximum difference of degree, that is, outdegree
	 * minus indegree.
	 * 
	 * @return
	 */
	public static Node getMaximumDifferenceNode(Graph graph) {
		Node max = null;
		Iterator<Node> iter = graph.nodes();

		if (iter.hasNext()) {
			max = iter.next();
			while (iter.hasNext()) {
				Node current = iter.next();
				if (current.getOutDegree() - current.getInDegree() > max.getOutDegree() - max.getInDegree()) {
					max = current;
				}
			}
		}

		return max;
	}
	
	/**
	 * Reverses and edge, that is, changes the source and the target and puts in
	 * the other way around. This method removes the old edge and creates a new
	 * one.
	 * 
	 * @param e
	 *            Edge to reverse
	 * @return The new reversed edge.
	 */
	public static Edge reverseEdge(Graph graph, Edge e) {
		Node source = e.getSourceNode();
		Node target = e.getTargetNode();
		graph.removeEdge(e);
		if ((e = graph.getEdge(target, source)) != null)
		    return e;
		return graph.addEdge(target, source);
	}
}

